{
  Autor: Michael Springwald
  
  Erstellt am: Samstag den 02.Augst.2008
  Änderungen
  ---------------------------------------------------------------------
01| Sontag, 03.Augst.2008         | Donnerstag, 14.August.2008        |
  ---------------------------------------------------------------------
02| Samstag, 16.Augst.2008        | Mittwoch, 22.Oktober.2008         |
  ---------------------------------------------------------------------
03| Dienstag, 28.Oktober.2008     | Freitag, 31.Oktober.2008          |
  ---------------------------------------------------------------------
04| Samstag, 01.Nov.2008          | Montag, 03.Nov.2008               |
  ---------------------------------------------------------------------
05| Donnerstag, 20.Nov.2008       | Freitag, 21.Nov.2008              |
  ---------------------------------------------------------------------
06| Samstag, 22.Nov.2008          |                                   |
  ---------------------------------------------------------------------
07|                               |                                   |
  ---------------------------------------------------------------------
08|                               |                                   |
  ---------------------------------------------------------------------
09|                               |                                   |
  ---------------------------------------------------------------------
10|                               |                                   |
  ---------------------------------------------------------------------

}

unit uPLlinesList;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, contnrs, LCLProc,LCLType,LCLIntf;

type

  TPLStyleChange = ( SC_FONTCOLOR=1, SC_FONTNAME=2,
                     SC_FONTSIZE=3, SC_FONTSTYLE=4,
                     SC_BRUSHCOLOR=5, SC_BRUSHSTYLE=6,
                     SC_PENCOLOR=7, SC_PENSTYLE=8,
                     SC_PENWIDTH=9, SC_FORMCANVAS=10
                   ); // TPLStyleChange

  TPLOnStyleChange = procedure(const aStyleName:String; const aStyleChange:TPLStyleChange) of Object;

  { TPLStyleItem }
  TPLStyleItem = class
  private
    fName: String;
    fonStyleChange: TPLOnStyleChange;
  protected
  public
    // Bei False wird ein Event ausgelöst, bei True wird kein Event ausgelöst
    NoChangeEvent:Boolean;

    constructor Create;
    destructor Destroy; override;

    // Löst das Event onStyleChange aus
    procedure DoStyleChange(const aStyleName:String; const aStyleChange:TPLStyleChange);
    // Alle Eigenschaften vom Item werden auf das Canvas übertragen
    procedure ToCanvas(var aCanvas:TCanvas); Virtual; Abstract;
    // Alle Notwendigen Eigenschaften vom Canvas wird auf das Item übertragen
    procedure FromCanvas(const aCanvas:TCanvas); Virtual; Abstract;
    // Kopiert einen Style auf den Anderen
    procedure Assign(const StyleItem:TPLStyleItem); Virtual; Abstract;
  published
    property Name:String read fName write fName;
    // Wird ausgelöst sobald sich eine Eigenschaft ändert
    property onStyleChange:TPLOnStyleChange read fonStyleChange write fonStyleChange;
  end; // TPLStyleItem

  { TPlStyleTextItem }

  TPlStyleTextItem = class(TPLStyleItem)
  private
    fBrushColor: TColor;
    fBrushStyle: TBrushStyle;
    fFontColor: TColor;
    fFontName: String;
    fFontSize: Integer;
    FfontStyle: TFontStyles;
    fPenColor: TColor;
    fPenStyle: TPenStyle;
    fPenWidth: Integer;
  protected
    procedure SetBrushColor(const AValue: TColor); Virtual;
    procedure SetBrushStyle(const AValue: TBrushStyle); Virtual;
    procedure SetFontColor(const AValue: TColor); Virtual;
    procedure SetFontName(const AValue: String); Virtual;
    procedure SetFontSize(const AValue: Integer); Virtual;
    procedure SetFontStyle(const AValue: TFontStyles); Virtual;
    procedure SetPenColor(const AValue: TColor); Virtual;
    procedure SetPenStyle(const AValue: TPenStyle); Virtual;
    procedure SetPenWidth(const AValue: Integer); Virtual;
  public
    constructor Create;
    destructor Destroy; override;
    // Alle Eigenschaften vom Item werden auf das Canvas übertragen
    procedure ToCanvas(var aCanvas:TCanvas); Override;
    // Alle Notwendigen Eigenschaften vom Canvas wird auf das Item übertragen
    procedure FromCanvas(const aCanvas:TCanvas); Override;
    // Kopiert einen Style auf den Anderen
    procedure Assign(const StyleItem:TPLStyleItem); Override;
  published
    // Text Eigenschaften
    property FontColor:TColor read fFontColor write SetFontColor;
    property FontName:String read fFontName write SetFontName;
    property FontSize:Integer read fFontSize write SetFontSize;
    property FontStyle:TFontStyles read FfontStyle write SetFontStyle;
    // Hintergrund Eigenschaften
    property BrushColor:TColor read fBrushColor write SetBrushColor;
    property BrushStyle:TBrushStyle read fBrushStyle write SetBrushStyle;
    // Rand Eigenschaften
    property PenColor:TColor read fPenColor write SetPenColor;
    property PenStyle:TPenStyle read fPenStyle write SetPenStyle;
    property PenWidth:Integer read fPenWidth write SetPenWidth;
  end; // TPlStyleTextItem
  

  { TPLLineItems }
  TPLLineItems = class(TPlStyleTextItem)
  private
    fw: Integer;
  protected
    fsymbol: Char;
    fx: Integer;
  public
    constructor Create;
    destructor Destroy; override;
  published
    property symbol:Char read fsymbol write fsymbol;
    property x:Integer read fx write fx;
    property w:Integer read fw write fw;

  end; // TPLLineItems

  { TPlLine }
  TPlLine = class
  private
    fh: Integer;
    fy: Integer;
    function GetLineItem(index: Integer): TPLLineItems;
    function GetLineText: String;
    procedure SetLineText(const AValue: String);
  protected
    fLineItems:TObjectList;
  public
    Data:TObject;
    Canvas:TCanvas;
    DefaultTextStyle:TPlStyleTextItem;
    constructor Create;
    destructor Destroy; override;
    
    function Add:TPLLineItems; virtual;
    function Insert(const aIndex:Integer):TPLLineItems; virtual;
    procedure Delete(const aindex:Integer);
    procedure Clear;
    function Count:Integer;
    function r1:TRect;

    procedure AddText(const aText:String);
    procedure InsertText(const aText:String; const aindex:Integer);
    function HighWidth:Integer;
    function HighHeight:Integer;

    property LineText:String read GetLineText write SetLineText;
    property LineItem[index:Integer]:TPLLineItems read GetLineItem;
  published
    property y:Integer read fy write fy;
    property h:Integer read fh write fh;
  end; // TPlLine

  { TPLLineList }
  TPLLineList = class
  private
    fDefaultTextStyle: TPlStyleTextItem;
    FHotItemIndex: Integer;
    fHotSelectItem: TPlStyleTextItem;
    FItemIndex: Integer;
    fSelectItem: TPlStyleTextItem;
    function GetLineList(index: Integer): TPlLine;
    procedure SetFHotItemIndex(const AValue: Integer);
    procedure SetItemIndex(const AValue: Integer);
  protected
    fLineList:TObjectList;
  public
    Data:TObject;
    Canvas:TCanvas;
    // Wo der Text auf dem Canvas gezeichnet werden soll
    Left, Top, Width, Height:Integer;
    // Der Abstand der Zeilen zueinander
    spaceTop:Integer;
    Buffer:TBitMap;
    ViewRect:TRect;
    FirstLine, LastLine:Integer;
    constructor Create;
    destructor Destroy; override;
    
    function Add:TPlLine; Virtual;
    function Insert(const aIndex:Integer):TPlLine; Virtual;

    procedure AddLine(const aLineText:String);
    
    procedure Clear;
    procedure Delete(const aIndex:Integer);
    function GetCount:Integer;
    function GetTextHeight:Integer;

    procedure LoadFromTextFile(const aFileName:String);
    procedure SetViewRecTop(const aItemIndex:Integer);

    {
      Ermittelet anhand der X und Y Position(Die in der Linken oberen Ecke
      sind), einen Zeilenindex.
    }
    function GetIndexFromXY(const aX, aY:Integer):Integer;

    {
      Ermitteln die "wirkliche" höhe der angeben Zeilen
    }
    function GetHighLineHeight(aVon, aBis:Integer):Integer;

    { Zeichnet die Angebe Line neu auf dem Canvas
      Parameter Beschreibung:
      |------------|-----------------------------------------------------------|
      | aCanvas    | Auf diesem Canvas wird gezeichnet                         |
      |------------|-----------------------------------------------------------|
      | aLineIndex | Bei -1 wird von CaretY ausgegangen                        |
      |------------|-----------------------------------------------------------|
      | NoNewDraw  | Bei True zeichnet PaintLine die Zeile nicht neu, und die  |
      |            |  wird auch nicht ausgeben. Das wird eigentlich nur für    |
      |            |  Paint gebraucht                                          |
      |------------|-----------------------------------------------------------|
      | NoDrawCaret| Bei True wird der Cursor nicht gezeichnet                 |
      |------------------------------------------------------------------------|
    }
    procedure PaintLine(const aLineIndex:Integer = -1;
                        const NoNewDraw:Boolean=False;
                        const NoDrawCaret:Boolean = False;
                        const NoSelectedItem:Boolean = False
                       );
// ------------------------------------------------------------------------
//  Zeichnet alle linen auf dem Angaben Canvas
    procedure Paint;
// ------------------------------------------------------------------------
//  Wird beim PaintLine aufgerufen, hier könnten z.b. das Aussehen der Einträge
//  Beinflusst werden, was bei einer Listbox Praktisch ist
    procedure SetStyleLine(const aLineIndex:Integer; aItemRec:TRect); Virtual;
// ------------------------------------------------------------------------
//  Ändert die Größe des Buffers
    procedure SetBufferSize(const aWidth, aHeight:Integer);
// ------------------------------------------------------------------------
    property LineList[index:Integer]:TPlLine read GetLineList;

  published
    property DefaultTextStyle:TPlStyleTextItem read fDefaultTextStyle write fDefaultTextStyle;

    property SelectItem:TPlStyleTextItem read fSelectItem write fSelectItem;

    property HotSelectItem:TPlStyleTextItem read fHotSelectItem write fHotSelectItem;
// ------------------------------------------------------------------------
// Der Ausgewählte Eintrag
   property ItemIndex:Integer read FItemIndex write SetItemIndex;
// Ein zweiter Index, wird genutzt um anzuzeigen wo die Maus drüber ist
   property HotItemIndex:Integer read FHotItemIndex write SetFHotItemIndex;

  end; // TPLLineList

implementation

// -- TPLStyleItem --
constructor TPLStyleItem.Create;
begin
  inherited Create;
  fonStyleChange:=nil;
  NoChangeEvent:=False;
end; // TPLStyleItem.Create

destructor TPLStyleItem.Destroy;
begin
  inherited Destroy;
end; // TPLStyleItem.Destroy

procedure TPLStyleItem.DoStyleChange(const aStyleName: String;
  const aStyleChange: TPLStyleChange);
begin
  if (Assigned(onStyleChange)) and (not NoChangeEvent) then
    onStyleChange(aStyleName, aStyleChange);
end; // TPLStyleItem.DoStyleChange

// -- TPlStyleTextItem -
procedure TPlStyleTextItem.SetBrushColor(const AValue: TColor);
begin
  if fBrushColor=AValue then exit;
  fBrushColor:=AValue;
  DoStyleChange(Name, SC_BRUSHCOLOR);
end; // TPlStyleTextItem.SetBrushColor

procedure TPlStyleTextItem.SetBrushStyle(const AValue: TBrushStyle);
begin
  if fBrushStyle=AValue then exit;
  fBrushStyle:=AValue;
  DoStyleChange(Name, SC_BRUSHSTYLE);
end; // TPlStyleTextItem.SetBrushStyle

procedure TPlStyleTextItem.SetFontColor(const AValue: TColor);
begin
  if fFontColor=AValue then exit;
  fFontColor:=AValue;
  DoStyleChange(Name, SC_FONTCOLOR);
end; // TPlStyleTextItem.SetFontColor

procedure TPlStyleTextItem.SetFontName(const AValue: String);
begin
  if fFontName=AValue then exit;
  fFontName:=AValue;
  DoStyleChange(Name, SC_FONTNAME);
end; // TPlStyleTextItem.SetFontName

procedure TPlStyleTextItem.SetFontSize(const AValue: Integer);
begin
  if fFontSize=AValue then exit;
  fFontSize:=AValue;
  DoStyleChange(Name, SC_FONTSIZE);
end; // TPlStyleTextItem.SetFontSize

procedure TPlStyleTextItem.SetFontStyle(const AValue: TFontStyles);
begin
  if FfontStyle=AValue then exit;
  FfontStyle:=AValue;
  DoStyleChange(Name, SC_FONTSTYLE);
end; // TPlStyleTextItem.SetFontStyle

procedure TPlStyleTextItem.SetPenColor(const AValue: TColor);
begin
  if fPenColor=AValue then exit;
  fPenColor:=AValue;
  DoStyleChange(Name, SC_PENCOLOR);
end; // TPlStyleTextItem.SetPenColor

procedure TPlStyleTextItem.SetPenStyle(const AValue: TPenStyle);
begin
  if fPenStyle=AValue then exit;
  fPenStyle:=AValue;
  DoStyleChange(Name, SC_PENSTYLE);
end; // TPlStyleTextItem.SetPenStyle

procedure TPlStyleTextItem.SetPenWidth(const AValue: Integer);
begin
  if fPenWidth=AValue then exit;
  fPenWidth:=AValue;
  DoStyleChange(Name, SC_PENWIDTH);
end; // TPlStyleTextItem.SetPenWidth

constructor TPlStyleTextItem.Create;
begin
  inherited Create;
end; // TPlStyleTextItem.Create

destructor TPlStyleTextItem.Destroy;
begin
  inherited Destroy;
end; // TPlStyleTextItem.Destroy

procedure TPlStyleTextItem.ToCanvas(var aCanvas: TCanvas);
begin
  aCanvas.Font.Color:=FontColor;
  aCanvas.Font.Name:=FontName;
  aCanvas.Font.Style:=FontStyle;
  aCanvas.Font.Size:=FontSize;
  aCanvas.Brush.Color:=BrushColor;
  aCanvas.Brush.Style:=BrushStyle;
  aCanvas.Pen.Color:=PenColor;
  aCanvas.Pen.Style:=PenStyle;
  aCanvas.Pen.Width:=PenWidth;
end; // TPlStyleTextItem.ToCanvas

procedure TPlStyleTextItem.FromCanvas(const aCanvas: TCanvas);
begin
  NoChangeEvent:=True;
  fFontColor:=aCanvas.Font.Color;
  fFontStyle:=aCanvas.Font.Style;
  fFontSize:=aCanvas.Font.Size;
  fFontName:=aCanvas.Font.Name;
  fBrushColor:=aCanvas.Brush.Color;
  fBrushStyle:=aCanvas.Brush.Style;
  fPenColor:=aCanvas.Pen.Color;
  fPenStyle:=aCanvas.Pen.Style;
  fPenWidth:=aCanvas.Pen.Width;
  NoChangeEvent:=True;
  DoStyleChange(Name,SC_FORMCANVAS);
end; // TPlStyleTextItem.FromCanvas

procedure TPlStyleTextItem.Assign(const StyleItem: TPLStyleItem);
begin
  FontName:=TPlStyleTextItem(StyleItem).FontName;
  FontSize:=TPlStyleTextItem(StyleItem).FontSize;
  FontColor:=TPlStyleTextItem(StyleItem).FontColor;
  FontStyle:=TPlStyleTextItem(StyleItem).FontStyle;
  BrushColor:=TPlStyleTextItem(StyleItem).BrushColor;
  BrushStyle:=TPlStyleTextItem(StyleItem).BrushStyle;
  PenColor:=TPlStyleTextItem(StyleItem).PenColor;
  PenStyle:=TPlStyleTextItem(StyleItem).PenStyle;
  PenWidth:=TPlStyleTextItem(StyleItem).PenWidth;
  onStyleChange:=StyleItem.onStyleChange;
end; // TPlStyleTextItem.Assign

function TPLLineList.GetLineList(index: Integer): TPlLine;
begin
  result:=(fLineList[index] as TPlLine);
end; // TPLLineList.GetLineList

procedure TPLLineList.SetFHotItemIndex(const AValue: Integer);
begin
  if FHotItemIndex=AValue then exit;
  if FHotItemIndex > -1 then begin
    PaintLine(FHotItemIndex,False,False,True);
    FHotItemIndex:=-1;
  end;

  if AValue > -1 then begin
    FHotItemIndex:=AValue;
    PaintLine(FHotItemIndex);
  end;
end; // TPLLineList.SetFHotItemIndex

procedure TPLLineList.SetItemIndex(const AValue: Integer);
begin
  if FItemIndex=AValue then exit;
//  if FItemIndex > -1 then begin
  //  PaintLine(FItemIndex,False,False,True);
  //  FItemIndex:=-1;
//  end;

  if AValue > -1 then begin
    FItemIndex:=AValue;
//    PaintLine(FItemIndex);
  end;
  paint;
end; // TPLLineList.SetItemIndex


// -- TPLLineList --
constructor TPLLineList.Create;
begin
  inherited Create;
  FHotItemIndex:=-1;
  FirstLine:=-1;
  LastLine:=-1;
  FItemIndex:=-1;
  fLineList:=TObjectList.Create;
  Data:=nil;
  Buffer:=TBitMap.Create;

  DefaultTextStyle:=TPlStyleTextItem.Create;
  DefaultTextStyle.Name:='DefaultTextStyle';
  DefaultTextStyle.FontColor:=clBlack;
  DefaultTextStyle.FontName:='courier [adobe]';
  DefaultTextStyle.FontStyle:=[];
  DefaultTextStyle.FontSize:=14;
  DefaultTextStyle.BrushColor:=clWhite;
  DefaultTextStyle.BrushStyle:=bsSolid;
  DefaultTextStyle.PenColor:=clBlack;
  DefaultTextStyle.PenStyle:=psSolid;
  DefaultTextStyle.PenWidth:=1;

  SelectItem:=TPlStyleTextItem.Create;
  SelectItem.Name:='SelectItem';
  SelectItem.FontColor:=clWhite;
  SelectItem.FontName:='courier [adobe]';
  SelectItem.FontStyle:=[];
  SelectItem.FontSize:=14;
  SelectItem.BrushColor:=clBlue;
  SelectItem.BrushStyle:=bsSolid;
  SelectItem.PenColor:=clBlack;
  SelectItem.PenStyle:=psSolid;
  SelectItem.PenWidth:=1;

  fHotSelectItem:=TPlStyleTextItem.Create;
  fHotSelectItem.Assign(DefaultTextStyle);
  fHotSelectItem.Name:='HotSelectItem';
  fHotSelectItem.FontColor:=clRed;
  fHotSelectItem.FontName:='courier [adobe]';
  fHotSelectItem.FontStyle:=[fsUnderline];
  fHotSelectItem.FontSize:=14;
//  fHotSelectItem.BrushColor:=clBlue;
  fHotSelectItem.BrushStyle:=bsSolid;
  fHotSelectItem.PenColor:=clBlack;
  fHotSelectItem.PenStyle:=psSolid;
  fHotSelectItem.PenWidth:=1;

//  fHotSelectItem: TPlStyleTextItem;
  spaceTop:=2;
end; // TPLLineList.Create

destructor TPLLineList.Destroy;
begin
  inherited Destroy;
end; // TPLLineList.Destroy

function TPLLineList.Add: TPlLine;
var
  aLine:TPlLine;
begin
  aLine:=TPlLine.Create;
  aLine.Canvas:=Canvas;
  aLine.DefaultTextStyle:=DefaultTextStyle;
  fLineList.Add(aLine);
  result:=aLine;
end; // TPLLineList.Add

function TPLLineList.Insert(const aIndex: Integer): TPlLine;
var
  aLine:TPlLine;
begin
  aLine:=TPlLine.Create;
  aLine.Canvas:=Canvas;
  aLine.DefaultTextStyle:=DefaultTextStyle;
  fLineList.Insert(aIndex,aLine);
  result:=aLine;
end; // TPLLineList.Insert

procedure TPLLineList.AddLine(const aLineText: String);
//var
//  x:Integer;
begin
   Add.LineText:=aLineText;
//  for x:=1 to Length(aLineText) do begin
//  end; // for i
end; // TPLLineList.AddLine

procedure TPLLineList.Clear;
begin
  fLineList.Clear;
end; // TPLLineList.Clear

procedure TPLLineList.Delete(const aIndex: Integer);
begin
  fLineList.Delete(aIndex);
end; // TPLLineList.Delete

function TPLLineList.GetCount: Integer;
begin
  result:=fLineList.Count;
end; // TPLLineList.GetCount

function TPLLineList.GetTextHeight: Integer;
var
  i,ph:integer;
begin
  ph:=0;
  for i:=0 to GetCount-1 do begin
    ph:=ph+LineList[i].HighHeight+spaceTop;
  end; // for i
  result:=ph;
end; // TPLLineList.GetTextHeight

function TPLLineList.GetIndexFromXY(const aX, aY: Integer): Integer;
var
  i,z:integer;
  r1:TRect;
  pt:TPoint;
begin
  z:=-1;
  pt:=Point(ax,ay);
//  writeln(FirstLine,'\',LastLine);
  for i:=FirstLine to LastLine do begin
    r1.Left:=0;
    r1.Top:=LineList[i].y;
    r1.Right:=Buffer.Width;
    r1.Bottom:=r1.top+LineList[i].h;
    if PtInRect(r1,pt) then begin
      z:=i;
      break;
    end;
  end; // for i
  result:=z;
end; // TPLLineList.GetIndexFromXY

procedure TPLLineList.LoadFromTextFile(const aFileName: String);
var
  StrList:TStrings;
  i:Integer;
begin
  StrList:=TStringList.Create;
  StrList.LoadFromFile(aFileName);
  for i:=0 to StrList.Count-1 do begin
    add.AddText(StrList[i]);
  end; // for i
  FreeAndNil(StrList);
end; // TPLLineList.LoadFromTextFile

procedure TPLLineList.SetViewRecTop(const aItemIndex: Integer);
var
  i,h,h1,h2,y,i2,index:Integer;
  c:TCanvas;
begin
  i:=buffer.Height-LineList[aItemIndex].HighHeight;
  if (LineList[aItemIndex].y > -1) and (aItemIndex >FirstLine) and (aItemIndex <LastLine) then begin
    writeln('OK:',random(1000),'|',aItemIndex,'|',FirstLine,'|',i);
    exit;
  end;

  i2:=0;
  if abs(LastLine-aItemIndex) <=1 then i2:=1;
  if (i2=0) and (abs(FirstLine-aItemIndex) <=1) then i2:=2;
  if i2 = 0 then i2:=3;
  writeln('I2:',i2,'|aItemIndex:',aItemIndex);

  h:=LineList[aItemIndex].HighHeight;
  case i2 of
    1:ViewRect.Top:=ViewRect.Top+h;
    2:ViewRect.Top:=ViewRect.Top-h;
    3: begin
      ViewRect.Top:=0;h1:=0; index:=0;
{      If aItemIndex >-1 Then // den Index zuerst auf Gültigkeit prüfen
      Begin
        While Index < aItemIndex Do begin
          Inc(ViewRect.Top,LineList[Index].HighHeight+spaceTop); // Alle Zeilenhöhen aufsummieren
          inc(index);
        end;}
      for i:=0 to GetCount-1 do begin
        if i <= aItemIndex then begin
          h2:=LineList[i].HighHeight;
          if h1+h2+spaceTop > buffer.Height then begin
            ViewRect.Top:=ViewRect.Top+h1+h2;
            h1:=0;
          end
          else
            h1:=h1+h2+spaceTop;
        end
        else
          break;
      end; // for i
    end;
  end;
end; // TPLLineList.SetViewRecTop

function TPLLineList.GetHighLineHeight(aVon, aBis:Integer): Integer;
var
  i, z:Integer;
begin
  z:=0;
  if aVon > fLineList.Count-1 then aVon:=fLineList.Count-1;
  if (aVon >-1) and (aBis > -1) then begin
    for i:=aVon to aBis do begin
      z:=z+LineList[i].h;
    end; // for i
  end;
  result:=z;
end; // TPlLine.GetHighLineHeight

procedure TPLLineList.PaintLine(const aLineIndex: Integer;const NoNewDraw:Boolean;
                                const NoDrawCaret: Boolean; const NoSelectedItem:Boolean = False);
var
  x,px,ty,pw,ph,py,w:Integer;
  c:TCanvas;
  LineItem:TPLLineItems;
  r:TRect;
  s, aSel:Boolean;
begin
  c:=Buffer.Canvas;
  px:=left+5; ty:=top;
  with Buffer.Canvas do begin

   // s:=((ItemIndex <> aLineIndex) and (not NoSelectedItem)) or (NoSelectedItem);
//    if not s then
      s:=(ItemIndex <> aLineIndex) and (HotItemIndex <> aLineIndex) and (not NoSelectedItem) or (NoSelectedItem);

    aSel:=False;
    //writeln('TEST:',ItemIndex,'\',aLineIndex,'\',NoSelectedItem,'\',random(100));
    if not NoSelectedItem then begin
      if (not aSel) and (ItemIndex = aLineIndex) then begin
        aSel:=True;
        SelectItem.ToCanvas(c);
        fHotItemIndex:=-1;
      end
      else begin
        if (not aSel) and (ItemIndex <> HotItemIndex) and (HotItemIndex = aLineIndex) then begin
          aSel:=True;
          HotSelectItem.ToCanvas(c);
        end;
      end;
    end;

    if (not aSel) then begin
      DefaultTextStyle.ToCanvas(c);
    end;

    ph:=LineList[aLineIndex].h;
    py:=LineList[aLineIndex].y;
    ty:=py;
    FillRect(0,py,buffer.Width, py+ph+spaceTop);

    for x:=0 to LineList[aLineIndex].Count-1 do begin
      LineItem:=LineList[aLineIndex].LineItem[x];
      if s then
        LineItem.ToCanvas(c);
      LineItem.w:=c.TextWidth(LineItem.symbol);
      pw:=LineItem.w;
      r.Left:=px;
//      if abs(ph-c.TextHeight(upperCase(LineItem.symbol))) > 10 then
  //      r.Top:=(py+ph)-c.TextHeight(upperCase(LineItem.symbol))
    //  else
        r.Top:=py;
      r.Right:=r.Left+pw+pw;
      r.Bottom:=r.Top+ph+spaceTop;
      if s then Buffer.Canvas.FillRect(r);

      SetStyleLine(aLineIndex, r);
      Buffer.Canvas.TextOut(r.Left,r.Top,LineItem.symbol );
      px:=px+pw;
    end; // for x

    ph:=LineList[aLineIndex].h+spaceTop;
    py:=LineList[aLineIndex].y;
    px:=left; ty:=top;

    if not NoDrawCaret then begin
      BitBlt(Canvas.Handle,0,py,buffer.Width,py+ph,
             buffer.Canvas.Handle,0,py,SRCCOPY
      );
    end;
  end; // with
end; // TPLLineList.PaintLine

procedure TPLLineList.Paint;
var
  i,py,ph,ty:Integer;
  c:TCanvas;
begin
  c:=Buffer.Canvas;
  py:=top;
  with Buffer.Canvas, DefaultTextStyle do begin
    ToCanvas(c);
   // Brush.Color:=clWhite;
    Brush.Style:=bsSolid;
    FillRect(2,0,buffer.Width, buffer.Height);
    ty:=py;
    //BrushStyle:=bsClear;
    FirstLine:=-1; LastLine:=-1;
    for i:=0 to fLineList.Count-1 do begin
      if i = ItemIndex then begin
        SelectItem.ToCanvas(c);
        LineList[i].h:=c.TextHeight(LineList[i].LineText);
      end
      else
        LineList[i].h:=LineList[i].HighHeight;

      ph:=LineList[i].h;
      if (py >=ViewRect.Top) and (py <=ViewRect.Top+buffer.Height+spaceTop+ph) then begin
        if FirstLine = -1 then FirstLine:=i;
        LineList[i].y:=ty;
        PaintLine(i,True,True);

        ty:=ty+ph+spaceTop;
//        if ty+ph>= buffer.Height then
  //      LastLine:=i;

      end;
      py:=py+LineList[i].h+spaceTop;
      if (py >=ViewRect.Top+ViewRect.Bottom) and (LastLine= -1) then begin
        if LastLine = -1 then LastLine:=i;
      end;

    end;
    if LastLine = -1 then LastLine:=fLineList.Count-1;
    BitBlt(Canvas.Handle,
           0,
           0,
           Width,
           Height,
           Buffer.Canvas.Handle,
           0,
           0,
           SRCCOPY
           );
//    Canvas.Draw(1,1,Buffer);
  end; // wiht
end; // TPLLineList.Paint

procedure TPLLineList.SetStyleLine(const aLineIndex: Integer; aItemRec:TRect);
begin

end; // TPLLineList.SetStyleLine

procedure TPLLineList.SetBufferSize(const aWidth, aHeight: Integer);
begin
  Buffer.Width:=aWidth;
  Buffer.Height:=aHeight;
  Buffer.Canvas.FillRect(0,0,aWidth, aHeight);
  Width:=aWidth; Height:=aHeight;
  ViewRect:=Rect(0,0,Width,Height);
end; // TPLLineList.SetBufferSize

function TPlLine.GetLineItem(index: Integer): TPLLineItems;
begin
  result:=(fLineItems[index] as TPLLineItems);
end; // TPlLine.GetLineItem

function TPlLine.GetLineText: String;
var
  x:Integer;
  str:String;
begin
  str:='';
  for x:=0 to Count-1 do begin
    str:=str+LineItem[x].symbol;
  end; // for x
  result:=str;
end; // TPlLine.GetLineText

procedure TPlLine.SetLineText(const AValue: String);
begin
  Clear; AddText(AValue);
end; // TPlLine.SetLineText

// -- TPLine --
constructor TPlLine.Create;
begin
  inherited Create;
  fLineItems:=TObjectList.Create;
  fy:=-1; Data:=nil;
end; // TPlLine.Create

destructor TPlLine.Destroy;
begin
  inherited Destroy;
end; // TPlLine.Destroy

function TPlLine.Add: TPLLineItems;
var
  aLineItem:TPLLineItems;
begin
  aLineItem:=TPLLineItems.Create;
  aLineItem.Assign(DefaultTextStyle);
  fLineItems.Add(aLineItem);
  result:=aLineItem;
end; // TPlLine.Add

function TPlLine.Insert(const aIndex: Integer): TPLLineItems;
var
  aLineItem:TPLLineItems;
begin
  aLineItem:=TPLLineItems.Create;
  aLineItem.Assign(DefaultTextStyle);

  fLineItems.Insert(aIndex,aLineItem);
  result:=aLineItem;
end; // TPlLine.Insert

procedure TPlLine.Delete(const aindex: Integer);
begin
  fLineItems.Delete(aindex);
end; // TPlLine.Delete

procedure TPlLine.Clear;
begin
  fLineItems.Clear;
end; // TPlLine.Clear

function TPlLine.Count: Integer;
begin
  result:=fLineItems.Count;
end; // TPlLine.Count

function TPlLine.r1: TRect;
//Var
//  aR:TRect;
begin
{  aR.Left:=0;
  aR.Top:=;

  aR.Right:=;
  aR.Botom:=;}

end; // TPlLine.r1

procedure TPlLine.AddText(const aText: String);
var
  x:Integer;
begin
  for x:=1 to UTF8Length(aText) do begin
    Add.symbol:=aText[x];
  end; // for x
end; // TPlLine.AddText

procedure TPlLine.InsertText(const aText: String; const aindex: Integer);
var
  x:Integer;
begin
  for x:=1 to UTF8Length(aText) do begin
    Insert(aindex+(x-1)).symbol:=aText[x];
  end; // for x
end; // TPlLine.InsertText

function TPlLine.HighWidth: Integer;
var
  x,z,w:Integer;
begin
  z:=-1;
  for x:=0 to Count-1 do begin
    LineItem[x].ToCanvas(Canvas);
    w:=Canvas.TextWidth(LineItem[x].symbol);
    if z+w > z then z:=z+w;
  end; // for x
  result:=z;
end; // TPlLine.HighWidth

function TPlLine.HighHeight: Integer;
var
  x,z,h1:Integer;
begin
  z:=-1;
  for x:=0 to Count-1 do begin
    LineItem[x].ToCanvas(Canvas);
    h1:=Canvas.TextHeight(LineItem[x].symbol);
    if h1 > z then z:=h1;
  end; // for x
  result:=z;
end; // TPlLine.HighHeight

{ TPLLineItems }

constructor TPLLineItems.Create;
begin
  inherited Create;
  x:=-1; w:=-1;
end; // TPLLineItems.Create

destructor TPLLineItems.Destroy;
begin
  inherited Destroy;
end; // TPLLineItems.Destroy

end.
